import time
import uuid
import six

from mazepa.queue import Queue
from mazepa.job import Job, AllJobsIndicator

class Scheduler:
    def __init__(self, queue_name=None, completion_queue_name=None,
            queue_region=None, threads=16, max_task_batch=20000):
        self.queue = Queue(queue_name=queue_name,
                completion_queue_name=completion_queue_name,
                queue_region=queue_region,
                threads=threads)

        self.unfinished_jobs = {}
        self.finished_jobs   = {}
        self.max_task_batch = max_task_batch

    def all_jobs_finished(self):
        return bool(self.unfinished_jobs)

    def submit_ready_jobs(self):
        jobs_ready = self.queue.get_completed()
        if jobs_ready is not None:
            self.submit_jobs(jobs_ready)
        else:
            unsubmitted_jobs = self.queue.get_unsubmitted_jobs(self.unfinished_jobs)
            if unsubmitted_jobs is not None:
                self.submit_jobs(unsubmitted_jobs)

    def execute_until_completion(self, sleep_gap_sec=4):
        jobs_spec = AllJobsIndicator()
        self.submit_jobs(jobs_spec)

        while True:
            self.submit_ready_jobs()
            if not self.unfinished_jobs:
                break
            time.sleep(sleep_gap_sec)

    def submit_jobs(self, jobs_spec):
        tasks = []
        jobs_just_finished = []

        for job_name, job in six.iteritems(self.unfinished_jobs):
            this_job_tasks = []
            # if this job is flagged for execution
            if isinstance(jobs_spec, AllJobsIndicator) or \
                    job_name in jobs_spec:
                if isinstance(job, Job):
                    this_job_tasks = job.get_task_batch()
                    print ("Got {} tasks from job '{}'".format(len(this_job_tasks),
                            job_name))
                    if this_job_tasks == []:
                        jobs_just_finished.append(job_name)
                else:
                    try:
                        this_job_tasks = job()
                    except StopIteration:
                        print ("Job '{}' is done!")
                        jobs_just_finished.append(job_name)
                for t in this_job_tasks:
                    t.job_name = job_name

                tasks.extend(this_job_tasks)
                if len(tasks) > self.max_task_batch:
                    break

        # Move finished jobs to finished dict
        for job_name in jobs_just_finished:
            print ("Flagging job as FINISHED: '{}'".format(job_name))
            self.finished_jobs[job_name] = self.unfinished_jobs[job_name]
            del self.unfinished_jobs[job_name]

        if len(tasks) > 0:
            print ("Scheduling {} tasks..".format(len(tasks)))
            self.queue.submit_mazepa_tasks(tasks)
        else:
            print ("No tasks to submit!")


    def register_job(self, job, job_name=None):
        '''
        job must be either an istance of type mazepa.Job
        or a simple generator
        '''
        if not isinstance(job, Job):
           if not inspect.isgeneratorfunction(job) and \
                   not inspect.isgenerator(job):
                       raise Exception("Registering job that is \
                               neither Mazepa job nor a generator is\
                               not supported. Submitted job type: {}".format(
                                   type(job)
                                   ))

        if job_name is None:
            job_name = uuid.uuid1()
            while job_name in self.unfinished_jobs:
                job_name = uuid.uuid1()
        else:
            if job_name in self.unfinished_jobs:
                raise Exception('Unable to register task "{}": \
                        task with name "{}" is already registerd'.format(job_name, job_name))

        self.unfinished_jobs[job_name] = job

        return job_name

    def unregister_job(self, job_name):
        del self.unfinished_jobs[job_name]



import boto3
import tenacity
import taskqueue
import time
import uuid
import multiprocessing
import functools
import json
from collections import defaultdict

from taskqueue import RegisteredTask as TQRegisteredTask

import mazepa

retry = tenacity.retry(
  reraise=True,
  stop=tenacity.stop_after_attempt(4),
  wait=tenacity.wait_full_jitter(0.5, 60.0),
  )


# TQ wrapper. Theretically don't have to use TQ library, but it's nice
class MazepaTaskTQ(TQRegisteredTask):
    def __init__(self, task=None,
            task_spec=None,
            completion_queue_name=None,
            completion_queue_region=None,
            task_id=None,
            job_id=None
            ):

        if task is not None:
            task_spec = mazepa.serialize(task)
            task_id = str(uuid.uuid4())
            job_id = task.job_name
        super().__init__(task_spec=task_spec,
                task_id=task_id, job_id=job_id,
                completion_queue_name=completion_queue_name,
                completion_queue_region=completion_queue_region)

    def execute(self):
        task = mazepa.deserialize(self.task_spec)
        print (f"Starting execution of {type(task)}")
        ts = time.time()
        task.execute()
        te = time.time()
        print (f"Done! Execution time: {te - ts} seconds. ")
        if self.completion_queue_name is not None:
            if self.job_id is None:
                raise Exception("'job_id' property must be set "
                        "for each task")
            send_completion_report(
                    queue_name=self.completion_queue_name,
                    queue_region=self.completion_queue_region,
                    task_id=self.task_id,
                    job_id=self.job_id
                    )

def send_completion_report(queue_name, queue_region,
        task_id, job_id):
    completion_queue = taskqueue.TaskQueue(
                                 queue_name,
                                 region_name=queue_region,
                                 n_threads=0,
                                 green=False)
    message = [{
        'Id': 'completion_report',
        'MessageBody':json.dumps({
                "task_id": task_id,
                "job_id": job_id
        })
    }]
    send_message(completion_queue, message)


def send_message(q, message):
    msg_ack = q.api.sqs.send_message_batch(
            QueueUrl=q.api.qurl,
            Entries=message
        )

    success = False
    for i in range(20):
        if 'Successful' in msg_ack and \
            len(msg_ack['Successful']) > 0:
            success = True
            break
        else:
            msg_ack = q.api.sqs.send_message_batch(
                    QueueUrl=q.api.qurl,
                    Entries=message
                )

    if success == False:
        raise ValueError(f"Failed to send message {message} to "
                         f"queue {q.api.qurl}")


class Queue:
    def __init__(self, queue_name=None, completion_queue_name=None,
            threads=0, queue_region=None):
        self.threads = threads
        self.queue_name = queue_name
        self.queue_region = queue_region
        self.completion_queue_name = completion_queue_name
        self.completion_registry = None

        if queue_name is None:
            self.local_execution = True
        else:
            self.local_execution = False
            if queue_name.startswith('fq://'):
                self.queue_type = 'fq'
                self.queue = taskqueue.TaskQueue(queue_name, green=False)
            else:
                self.queue_type = 'sqs'
                self.queue = taskqueue.TaskQueue(queue_name,
                        region=queue_region, green=False)
                self.queue_boto = boto3.client('sqs',
                                               region_name=queue_region)
                self.queue_url = self.queue_boto.get_queue_url(
                        QueueName=self.queue_name)["QueueUrl"]

            if completion_queue_name is not None:
                if self.queue_type == 'fq':
                    raise NotImplementedError("Completion queue is not supported with sqs")

                self.completion_registry = None
                self.completion_queue = \
                        taskqueue.TaskQueue(
                                completion_queue_name,
                                region=queue_region,
                                green=False)
                self.completion_queue_url = \
                        self.queue_boto.get_queue_url(
                              QueueName=
                                  completion_queue_name)["QueueUrl"]

                self.completion_registry = defaultdict(lambda: {})

            else:
                self.completion_queue = None
                self.completion_queue_url = None

    def submit_tq_tasks(self, tq_tasks):
        if self.completion_queue is not None:
            for t in tq_tasks:
                self.completion_registry[t.job_id][t.task_id] = True

        self.queue.insert(tq_tasks, parallel=self.threads)

    def submit_mazepa_tasks(self, mazepa_tasks):
        if self.local_execution:
            print ("Starting local task execution...")
            for t in mazepa_tasks:
                t.execute()
        else:
            pool = multiprocessing.Pool(self.threads)
            task_constructor = functools.partial(
                    MazepaTaskTQ,
                    completion_queue_name=self.completion_queue_name,
                    completion_queue_region=self.queue_region
                    )
            tq_tasks = pool.map(task_constructor, mazepa_tasks)
            self.submit_tq_tasks(tq_tasks)

    @retry
    def remote_queue_is_empty(self):
        """Is our remote queue empty?
        """
        if self.queue_type == 'sqs':
            attribute_names = ['ApproximateNumberOfMessages', 'ApproximateNumberOfMessagesNotVisible']
            is_empty = True
            for i in range(10):
                response = self.queue_boto.get_queue_attributes(
                        QueueUrl=self.queue_url,
                        AttributeNames=attribute_names)
                for a in attribute_names:
                    num_messages = int(response['Attributes'][a])
                    if num_messages > 0:
                        is_empty = False
                        break
                if i < 9:
                    time.sleep(0.5)

            return is_empty
        elif self.queue_type == 'fq':
            return self.queue.is_empty()
        else:
            raise Exception(f"Unsupported queue type: '{self.queue_type}'")

    def is_local_queue(self):
        return self.local_execution

    def is_remote_queue(self):
        return not self.is_local_queue()

    def get_completed(self):
        if self.is_local_queue():
            return mazepa.job.AllJobsIndicator()
        else:
            if self.completion_queue is None:
                if self.remote_queue_is_empty():
                    return mazepa.job.AllJobsIndicator()
                else:
                    time.sleep(5)
                    return None
            else:
                completed_jobs = self.check_completion_report()
                if completed_jobs is not None:
                    return completed_jobs
                else:
                    if self.remote_queue_is_empty():
                        return mazepa.job.AllJobsIndicator()
                    else:
                        time.sleep(5)
                        return None


    def check_completion_report(self):
        print ("Checking completion report...")
        completed_jobs = []
        while True:
            resp = self.queue_boto.receive_message(
                            QueueUrl=self.completion_queue_url,
                            AttributeNames=['All'],
                            MaxNumberOfMessages=10)

            if 'Messages' not in resp:
                break
            else:
                entries = []
                completed_tasks = []

                for i in range(len(resp['Messages'])):
                    entries.append({
                        'ReceiptHandle':\
                            resp['Messages'][i]['ReceiptHandle'],
                        'Id': str(i)
                    })
                    completed_tasks.append(json.loads(
                            resp['Messages'][i]['Body']))

                self.queue_boto.delete_message_batch(
                        QueueUrl=self.completion_queue_url,
                        Entries=entries)
            for t in completed_tasks:
                if t['job_id'] in self.completion_registry and \
                        t['task_id'] in self.completion_registry[t['job_id']]:
                    print (f"Deleting task {t['task_id']} from job {t['job_id']}")
                    del self.completion_registry[t['job_id']][t['task_id']]

                    if len(self.completion_registry[t['job_id']]) == 0:
                        completed_jobs.append(t['job_id'])
                        del self.completion_registry[t['job_id']]
                else:
                    print (f"Unregistered task {t['task_id']} from job {t['job_id']}")


        if len(completed_jobs) > 0:
            return completed_jobs
        else:
            return None

    def get_unsubmitted_jobs(self, full_job_list):
        unsubmitted_jobs = []
        if self.completion_queue is not None:
            for j in full_job_list:
                if j not in self.completion_registry:
                    unsubmitted_jobs.append(j)
        return unsubmitted_jobs

    def poll_tasks(self, lease_seconds):
        assert self.is_remote_queue
        self.queue.poll(lease_seconds=lease_seconds)



import types

class Task:
    def __init__(self, *kargs, **kwargs):
        self.job_id = None

    def __call__(self, *kargs, **kwargs):
        raise NotImplementedError


class Job:
    def __init__(self, *kargs, task_batch_size=300000, **kwargs):
        self.task_batch_size = task_batch_size
        self.task_generator = self.task_generator()
        self.task_batch_generator = self.task_batch_generator()

    def get_tasks(self):
        return next(self.task_generator)

    def task_generator(self):
        raise NotImplemented("Jobs must implement 'task_generator' function")

    def get_task_batch(self):
        return next(self.task_batch_generator)

    def task_batch_generator(self):
        '''
        yields: list of tasks that can be completed. when done, yields an empty list
        '''
        result = []
        while True:
            #squeeze the job until it's either done or returns a barrier
            try:
                task_batch = self.get_tasks()
                if task_batch is Barrier:
                    if len(result) == 0:
                        raise Exception(f"Job '{type(self)}' issued two Barriers in a row, "
                                         "or issued a barrier as the first task")
                    else:
                        yield result
                        result = []
                elif isinstance(task_batch, list):
                    result.extend(task_batch)
                elif isinstance(task_batch, types.GeneratorType):
                    for new_task in task_batch:
                        result.append(new_task)
                        if len(result) >= self.task_batch_size:
                            yield result
                            result = []
                else:
                    raise Exception(f"Object of unsupported type '{type(task_batch)}' yielded task_generator")

                if len(result) >= self.task_batch_size:
                    yield result
                    result = []

            except StopIteration:
                #means that this job has no more tasks, it's done
                yield result
                break


        yield []

class MazepaExecutionSignal:
    def __init__(self):
        pass

class Barrier(MazepaExecutionSignal):
    def __init__(self):
        pass

class JobReadyIndicator:
    def __init__(self):
        pass

class JobReadyList(JobReadyIndicator):
    def __init__(self, job_list):
        super().__init__()
        self.job_list = job_list

class AllJobsIndicator(JobReadyIndicator):
    def __init__(self):
        super().__init__()




import time
import uuid
import six

from mazepa.queue import Queue
from mazepa.job import Job, AllJobsIndicator

class Executor:
    def __init__(self, queue_name=None, queue_region=None,
            completion_queue_name=None):
        self.queue = Queue(queue_name=queue_name,
                completion_queue_name=completion_queue_name,
                queue_region=queue_region)

    def execute(self, lease_seconds):
        self.queue.poll_tasks(lease_seconds=lease_seconds)




from .job import Job, Task, Barrier
from .scheduler import Scheduler
from .serialization import serialize, deserialize
from .executor import Executor
from .click_interface import click_options,\
        parse_scheduler_from_kwargs, \
        parse_executor_from_kwargs




import boto3
import click

from mazepa.scheduler import Scheduler
from mazepa.executor import Executor

def click_options(cls):
    queue = click.option('--queue_name', '-q', nargs=1,
            type=str, default=None,
            help="Name of the queue where the tasks will be pushed to. "
            "For file queue, use 'fq://{path_to_shared_storage}' format."
            "If no 'fq://' prefix is given, the queue is assumed to be SQS. "
            "If not specified, tasks will be executed locally.")

    completion_queue = click.option('--completion_queue_name', nargs=1,
            type=str, default=None,
            help="Name of AWS SQS queue where completion of tasks will "
            "be reported. Must be distinct from the 'queue_name'. "
            "Providing a completion queue will speed up execution when "
            "running workers on unreliable machines (preemptible, spot).")

    region = click.option('--sqs_queue_region',  nargs=1,
            type=str, default='us-east-1',
            help="AWS region of  SQS queues. Task queue and completion "
            "queue must share the same region.")

    return queue(completion_queue(region(cls)))

def parse_queue_params(args):
    queue_name = args['queue_name']
    completion_queue_name = args['completion_queue_name']
    queue_region = args['sqs_queue_region']

    if queue_name is not None:
        s = boto3.Session()
        sqs_regions = s.get_available_regions('sqs')
        if queue_region not in sqs_regions:
            raise Exception(f"Invalid AWS SQS region '{queue_region}'. "
                f"Valid AWS SQS region list: {sqs_regions}")

            if queue_name is None and completion_queue_name is not None:
                raise Exception("'completion_queue_name' can only be "
                        "used when 'queue_name' is provided")
    return {
            'queue_name': queue_name,
            'completion_queue_name': completion_queue_name,
            'queue_region': queue_region
        }

def parse_scheduler_from_kwargs(args):
    queue_params = parse_queue_params(args)
    return Scheduler(**queue_params)

def parse_executor_from_kwargs(args):
    queue_params = parse_queue_params(args)
    return Executor(**queue_params)



import pickle
import codecs
import zlib

def serialize(obj):
    return codecs.encode(zlib.compress(pickle.dumps(obj, protocol=4)), "base64").decode()

def deserialize(s):
    return pickle.loads(zlib.decompress(codecs.decode(s.encode(), "base64")))
